/*
 * Copyright (C) 2010 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import com.google.appengine.api.datastore.DatastoreService;
import com.google.appengine.api.datastore.Entity;
import com.google.appengine.api.datastore.EntityNotFoundException;
import com.google.appengine.api.datastore.FetchOptions;
import com.google.appengine.api.datastore.Key;
import com.google.appengine.api.datastore.Query;
import com.google.appengine.api.datastore.Transaction;
import com.google.appengine.api.utils.SystemProperty;
import com.google.apphosting.api.ApiProxy;
import com.google.apphosting.api.DatastorePb;
import com.google.apphosting.api.ApiProxy.Environment;
import com.google.apphosting.api.DatastorePb.Schema;
import com.google.storage.onestore.v3.OnestoreEntity.EntityProto;
import com.google.storage.onestore.v3.OnestoreEntity.Path.Element;

/**
 * {@code AppEngineDatastoreService} is a convenient wrapper of 
 * {@code com.google.appengine.api.datastore.DatastoreService}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class AppEngineDatastoreService {

	private static Logger logger = Logger.getLogger(AppEngineDatastoreService.class.getName());
	
	private final DatastoreService datastore;
	
	/**
	 * Constructs a new {@code AppEngineDatastoreService} with the specified 
	 * {@code com.google.appengine.api.datastore.DatastoreService}.
	 * 
	 * @param datastore com.google.appengine.api.datastore.DatastoreService.
	 */
	public AppEngineDatastoreService(DatastoreService datastore) {
		this.datastore = datastore;
	}

	/**
	 * Gets App Engine DatastoreSession entity corresponding to the specified {@code Key}.
	 * If the entity doesn't exist in datastore, this method returns 
	 * {@code null}.
	 * 
	 * @param paramKey The key to get the entity.
	 * @return The entity corresponding to the specified {@code Key}.
	 */
	public Entity get(Key paramKey) {
		try {
			return datastore.get(paramKey);
		} catch (EntityNotFoundException e) {
			return null;
		}
	}
	
	/**
	 * Gets App Engine DatastoreSession entity corresponding to the specified 
	 * {@code Key} in the specified {@code Transaction}.
	 * 
	 * @param paramTransaction The transaction to get the entity.
	 * @param paramKey The key to get the entity.
	 * @return The entity corresponding to the specified {@code Key} in the 
	 * specified {@code Transaction}.
	 */
	public Entity get(Transaction paramTransaction, Key paramKey) {
		try {
			return datastore.get(paramTransaction, paramKey);
		} catch (EntityNotFoundException e) {
			return null;
		}
	}

	/**
	 * Puts the specified App Engine DatastoreSession entity into datastore with the 
	 * specified {@code Transaction}.
	 * 
	 * @param paramTransaction The transaction to put the entity.
	 * @param paramEntity The entity to be put.
	 * @return The {@code Key} allocated to the stored entity.
	 */
	public Key put(Transaction paramTransaction, Entity paramEntity) {
		return datastore.put(paramTransaction, paramEntity);
	}
	
	/**
	 * Puts the specified App Engine DatastoreSession entities into datastore with the 
	 * specified {@code Transaction}.
	 * 
	 * @param paramTransaction The transaction to put the entities.
	 * @param paramIterable The entities to be put. 
	 * @return
	 */
	public List<Key> put(Transaction paramTransaction, Iterable<Entity> paramIterable) {
		return datastore.put(paramTransaction, paramIterable);
	}
	
	/**
	 * Deletes entities corresponding to the specified {@code Key}s from App 
	 * Engine DatastoreSession in the specified {@code Transaction}.
	 * 
	 * @param paramTransaction The transaction to delete entities.
	 * @param paramIterable The {@code Key}s corresponding to the entities to be 
	 * deleted.
	 */
	public void delete(Transaction paramTransaction, Iterable<Key> paramIterable) {
		datastore.delete(paramTransaction, paramIterable);
	}
	
	/**
	 * Deletes entities corresponding to the specified {@code Key}s from App 
	 * Engine DatastoreSession in the specified {@code Transaction}.
	 * 
	 * @param paramTransaction The transaction to delete entities.
	 * @param paramArrayOfKey The {@code Key}s corresponding to the entities to 
	 * be deleted.
	 */
	public void delete(Transaction paramTransaction, Key... paramArrayOfKey) {
		datastore.delete(paramTransaction, paramArrayOfKey);
	}
	
	/**
	 * Begins transaction and returns it.
	 * 
	 * @return Current transaction.
	 */
	public Transaction beginTransaction() {
		return datastore.beginTransaction();
	}
	
	/**
	 * Executes the specified query without offset qualification.
	 * 
	 * @param paramQuery The query to be executed.
	 * @return The query result as {@code List} view.
	 */
	public List<Entity> query(Query paramQuery) {
		return query(paramQuery, FetchOptions.Builder.withOffset(0));
	}
	
	/**
	 * Executes the specified query in the specified transaction without offset 
	 * qualification.
	 * 
	 * @param paramTransaction The transaction in which the specified query is 
	 * executed.
	 * @param paramQuery The query to be executed.
	 * @return The query result as {@code List} view.
	 */
	public List<Entity> query(Transaction paramTransaction, Query paramQuery) {
		return query(paramTransaction, paramQuery, FetchOptions.Builder.withOffset(0));
	}
	
	private static boolean local = true;
	
	static {
		try {
			if (SystemProperty.environment.value() 
					== SystemProperty.Environment.Value.Production) {
				local = false;
			}
			
		} catch (Exception e) {}
		
		if (local) {
			logger.fine("Acid House is running on local environment");
		} else {
			logger.fine("Acid House is running on production environment");
		}
	}
	
	/**
	 * Executes the specified query with the specified fetch options.
	 * This method simulates "Kindless ancestor query" in the local development 
	 * environment. See <a href="http://code.google.com/p/googleappengine/issues/detail?id=2196">
	 * here</a>. This method assumes that current environment is local if <code>
	 * com.google.appengine.tools.development.ApiProxyLocal</code> can be loaded.
	 * 
	 * @param paramQuery The query to be executed.
	 * @param fetchOptions The fetch options which the specified query is 
	 * executed with.
	 * @return The query result as {@code List} view.
	 */
	public List<Entity> query(Query paramQuery, FetchOptions fetchOptions) {
		if (paramQuery.getKind() == null) {
			// Kindless ancestor query does not perform in the development 
			// environment. 
			// See Issue 2196:	ancestor query does not work on development server
			// http://code.google.com/p/googleappengine/issues/detail?id=2196
			if (local) {
				return kindlessAncestorQuery(null, paramQuery, fetchOptions);
			}
		}
		
		return datastore.prepare(paramQuery).asList(fetchOptions);
	}
	
	/**
	 * Executes the specified query in the specified transaction with the 
	 * specified fetch options.
	 * This method simulates "Kindless ancestor query" in the local development 
	 * environment. See <a href="http://code.google.com/p/googleappengine/issues/detail?id=2196">
	 * here</a>. This method assumes that current environment is local if <code>
	 * com.google.appengine.tools.development.ApiProxyLocal</code> can be loaded.
	 * 
	 * @param paramTransaction The transaction in which the specified query is 
	 * executed.
	 * @param paramQuery The query to be executed.
	 * @param fetchOptions The fetch options which the specified query is 
	 * executed with.
	 * @return The query result as {@code List} view.
	 */
	public List<Entity> query(Transaction paramTransaction, Query paramQuery, FetchOptions fetchOptions) {
		if (paramQuery.getKind() == null) {
			if (local) {
				return kindlessAncestorQuery(paramTransaction, paramQuery, fetchOptions);
			}
		}
		
		return datastore.prepare(paramTransaction, paramQuery).asList(fetchOptions);
	}
	
	private List<Entity> kindlessAncestorQuery(Transaction paramTransaction, 
			Query paramQuery, FetchOptions fetchOptions) {
		Set<String> kinds = new HashSet<String>();
		Environment environment = ApiProxy.getCurrentEnvironment();
		Schema schema = new Schema();
		schema.mergeFrom(ApiProxy.makeSyncCall(
				"datastore_v3", 
				"GetSchema", 
				new DatastorePb.GetSchemaRequest().setApp(environment.getAppId()).toByteArray()));
		
		for (EntityProto proto : schema.kinds()) {
			List<Element> elements = proto.getKey().getPath().elements();
			kinds.add(elements.get(elements.size() - 1).getType());
		}
		
		List<Entity> entities = new ArrayList<Entity>();
		
		for (String kind : kinds) {
			if (paramTransaction == null) {
				entities.addAll(datastore.prepare(
						new Query(kind, paramQuery.getAncestor())).asList(
						FetchOptions.Builder.withOffset(0)));
			} else {
				entities.addAll(datastore.prepare(
						paramTransaction, 
						new Query(kind, paramQuery.getAncestor())).asList(
						FetchOptions.Builder.withOffset(0)));
			}
		}
		
		return entities;
	}
	
}
