/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.logging.Logger;

import org.eiichiro.acidhouse.IndoubtException;

/**
 * {@code AppEngineGlobalTransaction} is a Google App Engine Low-level Datastore 
 * API based implementation of {@code Transaction}.
 * This implementation supports the strong consistency with atomic commitment 
 * in multiple entity groups.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class AppEngineGlobalTransaction extends AppEngineTransaction {

	private final Logger logger = Logger.getLogger(getClass().getName());
	
	private final AppEngineCoordinator coordinator;
	
	private final AppEngineStrongDatastoreSession session;
	
	/**
	 * Constructs a new {@code AppEngineGlobalTransaction} with the specified 
	 * {@code AppEngineStrongDatastoreSession} and {@code AppEngineCoordinator}.
	 * 
	 * @param session {@code AppEngineStrongDatastoreSession} that starts this 
	 * transaction.
	 * @param coordinator {@code AppEngineCoordinator} that starts this 
	 * transaction.
	 */
	public AppEngineGlobalTransaction(AppEngineStrongDatastoreSession session,
			AppEngineCoordinator coordinator) {
		super(session);
		this.coordinator = coordinator;
		this.session = session;
	}
	
	/**
	 * Constructs a new {@code AppEngineGlobalTransaction} with the specified 
	 * transaction id, {@code AppEngineStrongDatastoreSession} and 
	 * {@code AppEngineCoordinator}.
	 * 
	 * @param id The transaction id.
	 * @param session {@code AppEngineStrongDatastoreSession} that starts this 
	 * transaction.
	 * @param coordinator {@code AppEngineCoordinator} that starts this 
	 * transaction.
	 */
	public AppEngineGlobalTransaction(String id, 
			AppEngineStrongDatastoreSession session,
			AppEngineCoordinator coordinator) {
		super(id, session);
		this.coordinator = coordinator;
		this.session = session;
	}

	/** Commits current transaction with {@code AppEngineCoordinator}. */
	@Override
	public void commit() throws IndoubtException {
		session.coordinator.remove();
		coordinator.commit();
		logger.info("Transaction ["
				+ coordinator.getCurrentTransaction().getId()
				+ "] committed");
	}

	/** Rolls back current transaction with {@code AppEngineCoordinator}. */
	@Override
	public void rollback() {
		session.coordinator.remove();
		coordinator.rollback();
	}

}
